/*-----------------------------------------------------------------
  All code is provided for EDUCATIONAL purposes ONLY. There is no
  warranty and no support.

  main.c

  Garage door monitor - 
  PA7 monitors a signal indicating if the garage door is open.
  If it stays open too long a signal output on PA6 makes the
  garage door close. The 2 digit LED display shows how often the
  program closed the door. In addition there is a hold button to 
  override the program's function, a clear button to zero
  the count, and a buzzer to warn when the door is closing.

  Processor: ATtiny461
      Clock: Internal, 4.0 MHz

  jb20100308 - If the hold button is pressed when the beeper was
			   sounding, the beeper would not go off.
-----------------------------------------------------------------*/
#include <avr/io.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <avr/wdt.h>
#include <util/delay.h>
#include "common.h"
#include "led.h"

#define INIT_CTR0L   	0xb0	// 65536-50000 = 15536d or 3CB0h
#define INIT_CTR0H   	0x3c
#define INIT_CTR1	 	6		// 256 - 250

#define STATE_MONITOR						0
#define STATE_VERIFY_DOOR_OPEN				1
#define STATE_DOOR_OPEN						2
#define STATE_DOOR_OPEN_HOLD				3
#define STATE_WARNING_ON					4
#define	STATE_WARNING_OFF					5
#define STATE_VERIFY_DOOR_CLOSED			6
#define STATE_CLOSE_DOOR_INIT				10
#define STATE_CLOSE_DOOR					11
#define STATE_FAULT							15

#define EVENT_HOLDSW_RELEASED		1
#define EVENT_CLEARSW_RELEASED		2

#define NUMBER_WARNING_BEEPS		5
#define DOOR_OPEN_DELAY_MIN			10
#define MAX_CLOSINGS_TODAY			10

#define MS_IN_A_DAY				864000

// Global variables

volatile uint8_t	 ucCounter=0;			// activation count
volatile char cDisplay[2] = { 0, 0 };		// digits to display

static volatile uint8_t	 ucDoorOpen=0;			// door open flag
static volatile uint16_t uiDownCtr=0;			// for timing intervals
static volatile uint32_t uiDayCtr=MS_IN_A_DAY;	// for measuring days
static volatile uint8_t	 ucClosingsToday=0;		// system fault
static volatile uint8_t	 ucEventFlags=0;		// event flags

// Internal Functions

static void system_init(void);

/*-----------------------------------------------------------------
  main()

  The mainline is implemented as a state machine. External events
  make the state change to produce the desired outputs.
-----------------------------------------------------------------*/
int main(void)
{
  uint8_t		ucState=STATE_MONITOR;
  uint8_t		ucBeeps=0;

  system_init();			// initialise processor
  ucCounter = eeprom_read_byte( 0 );
  if ( ucCounter > 254 )
  { 
    ucCounter = 0;
	eeprom_write_byte( 0, ucCounter );
  }
  format_counter();			// update the display

  wdt_enable( WDTO_60MS );

  while( 1 )
  {
    switch ( ucState )
	{
	  case STATE_MONITOR:
		// We spend most of our time here, waiting for the door to open.
		if ( !uiDayCtr )
		{
		  // The day counter has expired. If the number of closings
		  // today reached the threshold, the program will be stuck in
		  // fault state so this won't happen here. If the number of
		  // closings is less than the threshold, the closings counter
		  // can safely be reset for another day. (Once fault
		  // mode has been entered, everything goes into suspense.)
		  ucClosingsToday = 0;
		  uiDayCtr = MS_IN_A_DAY;
		}

		if ( ucDoorOpen )
		{
		  clear_bit( ucEventFlags, EVENT_HOLDSW_RELEASED );
		  uiDownCtr = 50;				// 50 x 100ms = 5 sec.
		  ucState = STATE_VERIFY_DOOR_OPEN;
		  break;
		}

		if ( test_bit( ucEventFlags, EVENT_CLEARSW_RELEASED ) )
		{
		  // Reset the counter to zero
		  ucCounter = 0;
		  ucClosingsToday = 0;
		  eeprom_write_byte( 0, ucCounter );
		  format_counter();			// update the display
		  clear_bit( ucEventFlags, EVENT_CLEARSW_RELEASED );
		}

		_delay_loop_2(1000);			// waste 1ms
		break;

	  case STATE_VERIFY_DOOR_OPEN:
	    // Make sure the door stays open for 5 seconds. The down counter
		// has been seeded with 5 seconds worth of 8ms counts.
		if ( uiDownCtr )				// still counting down...
		{
		  _delay_loop_2(1000);			// waste 1ms
		  break;						// keep waiting
		}

		if ( ucDoorOpen )				// the door still open?
		{
		  uiDownCtr = DOOR_OPEN_DELAY_MIN * 600;	// DOOR_OPEN_DELAY_MIN x 60 x 100ms
		  ucState = STATE_DOOR_OPEN;	// wait for door to close
		}
		else
		{
		  uiDownCtr = 0;
		  clear_bit( ucEventFlags, EVENT_HOLDSW_RELEASED );
		  clear_bit( ucEventFlags, EVENT_CLEARSW_RELEASED );
		  ucState = STATE_MONITOR;		// go back to monitoring
		}
		break;

	  case STATE_DOOR_OPEN:
	    // Now we're waiting for the door to close or the interval to expire.
		if ( test_bit( ucEventFlags, EVENT_HOLDSW_RELEASED ) )
		{
		  // The hold button has been pressed and released
		  uiDownCtr = 0;				// cancel the downcount
		  clear_bit( ucEventFlags, EVENT_HOLDSW_RELEASED );
		  set_bit( PORTB, PB5 );		// turn on hold LED
		  ucState = STATE_DOOR_OPEN_HOLD;
		  break;
		}

		if ( !ucDoorOpen )				// the door has closed?
		{
		  uiDownCtr = 50;				// 50 x 100ms = 5 sec.
		  ucState = STATE_VERIFY_DOOR_CLOSED;
		  break;
		}
		
		if ( uiDownCtr )				// still counting down
		  _delay_loop_2(1000);			// waste 1ms
		else							// interval has expired
		{
		  // Prepare to sound the alarm a few times before closing the door
		  ucBeeps = NUMBER_WARNING_BEEPS;	// set number of beeps
		  uiDownCtr = 20;				// 20 x 100ms = 2 sec.
		  set_bit( PORTB, PB6 );		// sound the alarm
		  ucState = STATE_WARNING_ON;
		}
	    break;		

	  case STATE_DOOR_OPEN_HOLD:
	    // We're holding now. If the hold button is pressed again, restart
		// the door open state. If the door is closed, go to monitoring.
		if ( !ucDoorOpen )				// door is closing
		{
		  uiDownCtr = 50;				// 50 x 100ms = 5 sec.
		  ucState = STATE_VERIFY_DOOR_CLOSED;
		  break;
		}

		if ( test_bit( ucEventFlags, EVENT_HOLDSW_RELEASED ) )
		{
		  // Pressing and releasing the hold button during hold cancels
		  // the hold. Then go back to actively watching the open door.
		  clear_bit( ucEventFlags, EVENT_HOLDSW_RELEASED );
		  clear_bit( PORTB, PB5 );		// turn off hold LED
		  uiDownCtr = 50;				// 50 x 100ms = 5 sec.
		  ucState = STATE_VERIFY_DOOR_OPEN;
		}
		else
		  _delay_loop_2(1000);			// waste 1ms
	    break;		

	  case STATE_VERIFY_DOOR_CLOSED:
	    // Make sure the door is really closed by waiting 5 seconds.
		if ( uiDownCtr )				// still waiting
		{
		  _delay_loop_2(1000);			// waste 1ms
		  break;
		}

		if ( ucDoorOpen )				// the door is open now
		{
		  // I may come here because I tried to close the
		  // door but failed for some reason.
		  uiDownCtr = 50;				// 50 x 100ms = 5 sec.
		  ucState = STATE_VERIFY_DOOR_OPEN;
		  break;
		}

		// OK door is closed.
		clear_bit( ucEventFlags, EVENT_HOLDSW_RELEASED );
		clear_bit( ucEventFlags, EVENT_CLEARSW_RELEASED );
		clear_bit( PORTB, PB5 );		// turn off hold LED
		ucState = STATE_MONITOR;
	    break;

	  case STATE_WARNING_ON:
		// The buzzer is on, wait till its time to turn it off.
		// Unless the door closed, go back to monitoring
		if ( !ucDoorOpen )				// door is closed
		{
		  clear_bit( PORTB, PB6 );		// silence the alarm
		  clear_bit( ucEventFlags, EVENT_HOLDSW_RELEASED );
		  clear_bit( ucEventFlags, EVENT_CLEARSW_RELEASED );
		  clear_bit( PORTB, PB5 );		// turn off hold LED
		  ucState = STATE_MONITOR;
		  break;
		}

		if ( test_bit( ucEventFlags, EVENT_HOLDSW_RELEASED ) )
		{
		  // the hold button has been pressed/released
		  uiDownCtr = 0;				// stop counting
		  clear_bit( ucEventFlags, EVENT_HOLDSW_RELEASED );
		  set_bit( PORTB, PB5 );		// turn on hold LED
		  clear_bit( PORTB, PB6 );		// jb20100308 silence the alarm
		  ucState = STATE_DOOR_OPEN_HOLD;
		  break;
		}

		if ( uiDownCtr )				// still waiting
		{
		  _delay_loop_2(1000);			// waste 1ms
		  break;
		}

		clear_bit( PORTB, PB6 );		// silence the alarm
		--ucBeeps;
		if ( !ucBeeps )
		{
		  // no more beeps
		  ucState = STATE_CLOSE_DOOR_INIT;
		  break;
		}

		uiDownCtr = 30;					// 30 x 100ms = 3 seconds
		ucState = STATE_WARNING_OFF;
	    break;		

	  case STATE_WARNING_OFF:
		// The buzzer is off, wait till its time to turn it on.
		// Unless the door closed, go back to monitoring
		if ( !ucDoorOpen )				// door is closed
		{
		  clear_bit( ucEventFlags, EVENT_HOLDSW_RELEASED );
		  clear_bit( ucEventFlags, EVENT_CLEARSW_RELEASED );
		  clear_bit( PORTB, PB5 );		// turn off hold LED
		  ucState = STATE_MONITOR;
		  break;
		}

		if ( test_bit( ucEventFlags, EVENT_HOLDSW_RELEASED ) )
		{
		  // the hold button has been pressed/released
		  uiDownCtr = 0;				// stop counting
		  clear_bit( ucEventFlags, EVENT_HOLDSW_RELEASED );
		  set_bit( PORTB, PB5 );		// turn on hold LED
		  ucState = STATE_DOOR_OPEN_HOLD;
		  break;
		}

		if ( uiDownCtr )				// still waiting
		{
		  _delay_loop_2(1000);			// waste 1ms
		  break;
		}

		set_bit( PORTB, PB6 );			// sound the alarm
		uiDownCtr = 20;					// 20 x 100ms = 2 seconds
		ucState = STATE_WARNING_ON;
	    break;
		
	  case STATE_CLOSE_DOOR_INIT:
		// If the door is still open, close it
		// Otherwise go back to monitoring
		if ( ucDoorOpen )				// door is still open
		{
		  ++ucCounter;					// record our action on the led display
		  ++ucClosingsToday;
		  format_counter();
		  eeprom_write_byte( 0, ucCounter );

		  if ( ucClosingsToday > MAX_CLOSINGS_TODAY )
		  {
		    uiDownCtr = 5;				// 5 x 100ms = 500ms
		    ucState = STATE_FAULT;
		  }
		  else
		  {
		    clear_bit( PORTA, PA6 );	// trigger close the door relay
		    uiDownCtr = 4;				// 4 x 100ms = 400ms
		    ucState = STATE_CLOSE_DOOR;
		  }
		  break;
		}

		clear_bit( ucEventFlags, EVENT_HOLDSW_RELEASED );
		clear_bit( ucEventFlags, EVENT_CLEARSW_RELEASED );
		clear_bit( PORTB, PB5 );		// turn off hold LED
		ucState = STATE_MONITOR;
	    break;
		
	  case STATE_CLOSE_DOOR:
		// Waiting to release the door close relay
		if ( uiDownCtr )				// still waiting
		{
		  _delay_loop_2(1000);			// waste 1ms
		  break;
		}

		set_bit( PORTA, PA6 );			// release close the door relay
		uiDownCtr = 50;					// 50 x 100ms = 5 sec.
		ucState = STATE_VERIFY_DOOR_CLOSED;
	    break;
		
	  case STATE_FAULT:
	    // We've tried to close the door the maximum number of times
		// today. If there is a fault in the door, we don't want to
		// keep trying indefinitely. The only way out of here is to 
		// press the clear button or power cycle the unit.
		if ( test_bit( ucEventFlags, EVENT_CLEARSW_RELEASED ) )
		{
		  // In fault state pressing the clear button clears the
		  // fault, not the counter
		  ucClosingsToday = 0;
		  uiDownCtr = 0;
		  clear_bit( ucEventFlags, EVENT_CLEARSW_RELEASED );
		  clear_bit( ucEventFlags, EVENT_HOLDSW_RELEASED );
		  clear_bit( PORTB, PB5 );		// turn off hold LED
		  ucState = STATE_MONITOR;
		  break;
		}

		// To indicate fault we flash the Hold LED on and off.
		if ( uiDownCtr )				// still waiting
		{
		  _delay_loop_2(1000);			// waste 1ms
		  break;
		}

		set_bit( PINB, PB5 );			// toggle the hold LED
		uiDownCtr = 5;					// 5 x 100ms = 500ms
	    break;
	}				// end switch

	wdt_reset();
  }					// end while(1)
}

/*-----------------------------------------------------------------
  Timer 0 overflow interrupt handler - updates the down counter

  Since the clock is running at 4.0 mHz, the counter increments
  every 250.0 ns. Using a prescaler of 8 gives a cycle time of
  2us. So if we count to 50,000 we get 100ms interrupts.
-----------------------------------------------------------------*/
ISR(TIMER0_OVF_vect)
{
  // Maintain the down counter for interval timing

  if ( uiDownCtr )
    --uiDownCtr;

  if ( uiDayCtr )
    --uiDayCtr;

  TCNT0L = INIT_CTR0L;					// reinitialize counter
  TCNT0H = INIT_CTR0H;
}

/*-----------------------------------------------------------------
  Timer 1 overflow interrupt handler - tests for button presses
  and switches LED digits

  Since the clock is running at 4.0 mHz, the counter increments
  every 250 ns. Using a prescaler of 128 gives us 32us. If we count
  250 times we get an interrupt every 8 ms.

  To achieve an interval of approximately 8 ms for selecting
  the next LED digit. By sequencing each digit fast enough, the
  LED digits both appear to be on.

  Test the buttons every interrupt and store 8 samples each.
  When all the bits are 1 then the button is pressed.
-----------------------------------------------------------------*/
ISR(TIMER1_OVF_vect)
{
  static uint8_t ucIndex=0;
  static uint8_t ucMask=1;
  static uint8_t ucHoldSw=0, ucHoldSwPrev=0;
  static uint8_t ucClearSw=0, ucClearSwPrev=0;
  static uint8_t ucOpenSw=0;

  // Select and display the next LED digit

  ledSelectDigit( NO_DIGIT );               // turn all digits off

  ledOutputNumber( cDisplay[ucIndex] );		// output the number

  ledSelectDigit( ucIndex );                // select the digit

  ++ucIndex;
  if ( ucIndex == 2 ) ucIndex = 0;

  // Read current input states

  if ( test_bit( PINB, PB3 ) )		// switch off
    ucHoldSw &= ~ucMask;
  else								// switch on
    ucHoldSw |= ucMask;

  if ( test_bit( PINB, PB4 ) )		// switch off
    ucClearSw &= ~ucMask;
  else								// switch on
    ucClearSw |= ucMask;

  if ( test_bit( PINA, PA7 ) )		// door open
    ucOpenSw |= ucMask;				
  else								// door closed
    ucOpenSw &= ~ucMask;

  ucMask = ucMask << 1;
  if ( !ucMask ) ucMask = 1;

  // Create button release events

  if ( ucHoldSw == 0xff )
  {
    // The hold switch is closed now
	if ( !ucHoldSwPrev )
	  // The hold switch was open before
	  ucHoldSwPrev = ucHoldSw;		// remember the button has just been pressed
  }
  else
    if ( !ucHoldSw )
	{
	  // The hold switch is open now
	  if ( ucHoldSwPrev == 0xff )
	  {
	    // The hold switch was closed before
		ucHoldSwPrev = ucHoldSw;	// remember the button has just been released
		set_bit( ucEventFlags, EVENT_HOLDSW_RELEASED );
	  }
	}

  if ( ucClearSw == 0xff )
  {
    // The clear switch is closed now
	if ( !ucClearSwPrev )
	  // The clear switch was open before
	  ucClearSwPrev = ucClearSw;		// remember the button has just been pressed
  }
  else
    if ( !ucClearSw )
	{
	  // The clear switch is open now
	  if ( ucClearSwPrev == 0xff )
	  {
	    // The clear switch was closed before
		ucClearSwPrev = ucClearSw;	// remember the button has just been pressed
		set_bit( ucEventFlags, EVENT_CLEARSW_RELEASED );
	  }
	}

  // Look for garage door opening/closing

  if ( ucOpenSw == 0xff )
  {
    // The door is open now
	if ( !ucDoorOpen )
	  // The door was closed before
	  ucDoorOpen = 1;
  }
  else
    if ( !ucOpenSw )
	{
	  // The door is closed now
	  if ( ucDoorOpen )
	    // The door was open before
		ucDoorOpen = 0;
	}

  TCNT1 = INIT_CTR1;			// reinitialize the timer
}
/*-----------------------------------------------------------------
  system_init()

  PORTB: PB0 (input) MOSI (ISP)
         PB1 (input) MISO (ISP)
         PB2 (input) SCK (ISP)
         PB3 (input) hold button
         PB4 (input) reset button
         PB5 (output) Hold LED 
         PB6 (output) Buzzer
         PB7 (input) !Reset (ISP)

  PORTA: PA0 (output) LED BCD "A"
         PA1 (output) LED BCD "B"
         PA2 (output) LED BCD "C" 
         PA3 (output) LED BCD "D" 
         PA4 (output) LED digit 1 
         PA5 (output) LED digit 2
         PA6 (output) relay 
         PA7 (input) door open signal

-----------------------------------------------------------------*/
static void system_init()
{
  // The fuses are set to provide an internal clock, running at
  // 8.0 MHz with a prescaler of 8. Change the prescaler from
  // 8 to 2. The procedure requires two steps as per the device
  // specifications.
  CLKPR = 0b10000000;       // set clock prescaler change enable
  CLKPR = 0b00000001;       // set prescaler to 2

  DDRB  = 0b01100000;       // PortB: all input except 5, 6
  PORTB = 0b10011111;       // initialise PortB, pullup 0..4, 7

  DDRA  = 0b01111111;       // PortA: all output except 7
  PORTA = 0b01000000;       // initialise PortA (PA6 high)

  // Initialize Timer0. Timer0 is used to produce 100ms ticks
  // for various timing purposes.
  TCCR0A = 0b10000000;       // normal 16-bit counter mode
  TCCR0B = 0b00000010;       // prescaler = 8 (turns timer0 on)
  TCNT0L = INIT_CTR0L;
  TCNT0H = INIT_CTR0H;
  set_bit( TIMSK, TOIE0 );	// enable timer/counter0 overflow interrupt

  // Initialize Timer1. Timer1 is used to debounce button presses
  // and switch LED digits every 8 ms.
  TCCR1A = 0b00000000;		// normal operation
  TCCR1B = 0b00001000;		// prescaler = 128 (turns timer1 on)
  TCNT1 = INIT_CTR1;
  set_bit( TIMSK, TOIE1 );	// enable timer/counter1 overflow interrupt

  sei();                    // enable global interrupts
}

